import math
import time
import rs2py
import datetime
from decimal import Decimal
from typing import List, Union


def price_ceil(price: float) -> float:
    logp = math.floor(math.log10(price))
    quinary = ((price / 10**logp) // 5) if logp >= 1 else 1
    n = min(10 ** (3 - logp - quinary), 100)
    return round(math.ceil(price * n + (5 - (price * n % 5)) * (1 - quinary)) / n, 2)


def price_floor(price: float) -> float:
    logp = math.floor(math.log10(price))
    quinary = ((price / 10**logp) // 5) if logp >= 1 else 1
    n = min(10 ** (3 - logp - quinary), 100)
    return round(math.floor(price * n - ((price * n % 5) * (1 - quinary))) / n, 2)


def price_round(price: float, up: bool = False) -> float:
    roudnf = math.ceil if up else math.floor
    price = Decimal(f"{price}")
    logp = math.floor(math.log10(price))
    quinary = ((price / 10**logp) // 5) if logp >= 1 else 1
    n = min(10 ** (3 - logp - quinary), 100)
    return float(
        round(
            roudnf(price * n + ((5 * int(up) - (price * n % 5)) * (1 - quinary))) / n, 2
        )
    )


def price_limit(price: float, up: float, down: float):
    if price > up:
        return up
    elif price < down:
        return down
    return price


def price_move(price: float, tick: int) -> float:
    return rs2py.get_price_tick_move(price, tick)


def price_between_tick(p0: float, p1: float) -> int:
    return rs2py.get_price_between_tick(p0, p1)


def quantity_num_split(quantity: int, num: int) -> List[int]:
    neg = 1 if quantity > 0 else -1
    quantity_remain = abs(quantity)
    quantity_split_res = [(quantity_remain // num) * neg] * num
    quantity_remain -= abs(sum(quantity_split_res))
    for idx, _ in enumerate(quantity_split_res):
        qadd = math.ceil(quantity_remain / (num - idx))
        quantity_remain -= qadd
        quantity_split_res[idx] += qadd * neg
        if not quantity_remain:
            break
    return quantity_split_res


def quantity_split(quantity: float, threshold: int) -> List[int]:
    return [threshold if quantity > 0 else -threshold] * (
        abs(quantity) // threshold
    ) + [
        (abs(quantity) % threshold) * (1 if quantity > 0 else -1),
    ]


def sleep_until(t: Union[datetime.time, tuple]) -> None:
    if isinstance(t, tuple):
        t = datetime.time(*t)
    now = datetime.datetime.utcnow() + datetime.timedelta(hours=8)
    d = datetime.timedelta(days=1) if now.hour > 13 else datetime.timedelta(days=0)
    until_time = (
        datetime.datetime(now.year, now.month, now.day, t.hour, t.minute, t.second) + d
    )
    delta = until_time - now
    delta_sec = delta.total_seconds()
    if delta_sec > 0:
        time.sleep(delta_sec)



import shioaji as sj
from typing import Dict, Optional
from loguru import logger
from shioaji.constant import TFTStockPriceType

from .io.file import read_position
from .utils import price_round, price_limit
from .position import Position, PriceSet
from .data import Snapshot


class StrategyBase:
    name: str

    def entry_positions(self):
        raise NotImplementedError()

    def cover_price_set(self, position: Position, snapshot: Optional[Snapshot] = None):
        raise NotImplementedError()

    def cover_positions(
        self, positions: Dict[str, Position], snapshots: Dict[str, Snapshot] = dict()
    ):
        raise NotImplementedError()

    def cover_price_set_onclose(self, position: Position):
        if position.status.open_quantity == 0:
            return []
        return [
            PriceSet(
                price=position.contract.limit_down
                if position.status.open_quantity > 0
                else position.contract.limit_up,
                quantity=position.status.open_quantity * -1,
                price_type=TFTStockPriceType.LMT,
            )
        ]

    def cover_positions_onclose(self, positions: Dict[str, Position]):
        for code, position in positions.items():
            position.cond.cover_price = self.cover_price_set_onclose(position)
        return positions


class StrategyBasic(StrategyBase):
    def __init__(
        self,
        entry_pct: float = 0.05,
        stop_loss_pct: float = 0.09,
        stop_profit_pct: float = 0.09,
        position_filepath: str = "position.txt",
        contracts: sj.contracts.Contracts = sj.contracts.Contracts(),
    ) -> None:
        self.position_filepath = position_filepath
        self.entry_pct = entry_pct
        self.stop_loss_pct = stop_loss_pct
        self.stop_profit_pct = stop_profit_pct
        self.contracts = contracts
        self.name = "dt1"
        self.read_position_func = read_position

    def entry_positions(self):
        positions = self.read_position_func(self.position_filepath)
        entry_args = []
        for code, pos in positions.items():
            contract = self.contracts.Stocks[code]
            if not contract:
                logger.warning(f"Code: {code} not exist in TW Stock.")
                continue
            stop_loss_price = contract.reference * (
                1 + (-1 if pos > 0 else 1) * (self.stop_loss_pct)
            )
            stop_loss_price = price_round(stop_loss_price, pos > 0)
            stop_loss_price = price_limit(
                stop_loss_price, contract.limit_up, contract.limit_down
            )
            stop_profit_price = contract.reference * (
                1 + (1 if pos > 0 else -1) * (self.stop_profit_pct)
            )
            stop_profit_price = price_round(stop_profit_price, pos < 0)
            stop_profit_price = price_limit(
                stop_profit_price, contract.limit_up, contract.limit_down
            )
            entry_price = contract.reference * (
                1 + (-1 if pos > 0 else 1) * self.entry_pct
            )
            entry_price = price_round(entry_price, pos > 0)
            entry_price = price_limit(
                entry_price, contract.limit_up, contract.limit_down
            )
            entry_args.append(
                {
                    "code": code,
                    "pos": pos,
                    "entry_price": [
                        PriceSet(
                            price=entry_price,
                            quantity=pos,
                            price_type=TFTStockPriceType.LMT,
                        )
                    ],
                    "stop_profit_price": [
                        PriceSet(
                            price=stop_profit_price,
                            quantity=pos,
                            price_type=TFTStockPriceType.MKT,
                        )
                    ],
                    "stop_loss_price": [
                        PriceSet(
                            price=stop_loss_price,
                            quantity=pos,
                            price_type=TFTStockPriceType.MKT,
                        )
                    ],
                }
            )
        return entry_args

    def cover_price_set(self, position: Position, snapshot: Optional[Snapshot] = None):
        return self.cover_price_set_onclose(position)

    def cover_positions(
        self, positions: Dict[str, Position], snapshots: Dict[str, Snapshot] = dict()
    ):
        return self.cover_positions_onclose()



from typing import Union, Dict
from pathlib import Path


def read_position(filepath: Union[Path, str]) -> Dict[str, int]:
    p = Path(filepath)
    if p.exists() and p.is_file():
        content = p.read_text()
        return {
            r[0]: int(float(r[1]))
            for r in [l.split("\t") for l in content.split("\n") if l]
        }
    else:
        raise FileNotFoundError(f"filepath: '{filepath}' not exist.")


def read_csv_position(
    filepath: Union[Path, str], with_header: bool = True
) -> Dict[str, int]:
    p = Path(filepath)
    if p.exists() and p.is_file():
        content = p.read_text()
        return {
            r[0]: {
                "pos": int(float(r[1])),
                "stop_loss_tick": int(r[2]),
                "cover_pct": float(r[3]),
            }
            for r in [
                l.split(",") for l in content.split("\n")[int(with_header) :] if l
            ]
        }
    else:
        raise FileNotFoundError(f"filepath: '{filepath}' not exist.")



import time
import random
import datetime
from typing import Callable, Dict
from dataclasses import dataclass
from threading import Lock
from concurrent.futures import ThreadPoolExecutor
import xxhash
import shioaji as sj
from shioaji.constant import OrderState, Exchange, Action, TFTStockPriceType

from .data import Snapshot


class SimulationShioaji:
    def __init__(self, order_deal_handler: Callable[[OrderState, Dict], None]):
        self.order_callback = order_deal_handler
        self.executor = ThreadPoolExecutor(max_workers=24)
        self.use_chars = (
            "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
        )
        self.seqno_counter = 0
        self.snapshots: Dict[str, Snapshot] = {}
        self.lmt_price_trades: Dict[str, Dict[str, sj.order.Trade]] = {}
        self.lock = Lock()

    def quote_callback(self, exchange: Exchange, tick: sj.TickSTKv1):
        if not tick.simtrade:
            if tick.code in self.snapshots:
                s = self.snapshots[tick.code]
                s.price = tick.close
            else:
                self.snapshots[tick.code] = Snapshot(tick.close)
            with self.lock:
                if tick.code in self.lmt_price_trades:
                    lmt_price_trades = self.lmt_price_trades[tick.code]
                    pop_order_ids = []
                    for order_id, trade in lmt_price_trades.items():
                        if (
                            trade.order.action == Action.Buy
                            and tick.close <= trade.order.price
                        ) or (
                            trade.order.action == Action.Sell
                            and tick.close >= trade.order.price
                        ):
                            deal_msg = self.gen_deal_msg(
                                trade,
                                quantity=trade.order.quantity,
                                price=tick.close,
                            )
                            self.order_callback(OrderState.TFTDeal, deal_msg)
                            pop_order_ids.append(order_id)
                    for order_id in pop_order_ids:
                        self.lmt_price_trades[tick.code].pop(order_id)
                    if not lmt_price_trades:
                        self.lmt_price_trades.pop(tick.code)

    def place_order(
        self,
        contract: sj.contracts.Contract,
        order: sj.order.TFTStockOrder,
        timeout: int = 5000,
    ):
        trade = sj.order.Trade(
            contract,
            order,
            sj.order.OrderStatus(status=sj.order.Status.PreSubmitted),
        )
        future = self.executor.submit(self.call_order_callback, trade, "New")
        # future.result()
        return trade

    def cancel_order(self, trade: sj.order.Trade, timeout: int = 5000):
        future = self.executor.submit(self.call_order_callback, trade, "Cancel")
        future.result()
        return trade

    def update_status(
        self,
        account: sj.Account = None,
        trade: sj.order.Trade = None,
        timeout: int = 5000,
    ):
        pass

    def gen_order_msg(self, trade: sj.order.Trade, op_type: str):
        if op_type == "New":
            self.seqno_counter += 1
            trade.order.seqno = f"{self.seqno_counter:0>6}"
            trade.order.id = xxhash.xxh32_hexdigest(trade.order.seqno)
            trade.order.ordno = ("").join(random.sample(self.use_chars, 5))
            trade.status.status = sj.order.Status.Submitted
            op_code = "00"
        elif op_type == "Cancel":
            if trade.status.status == sj.order.Status.Filled:
                op_code = "11"
                cancel_quantity = 0
            else:
                op_code = "00"
                cancel_quantity = trade.order.quantity
                if trade.contract.code in self.lmt_price_trades:
                    trades = self.lmt_price_trades[trade.contract.code]
                    if trade.order.id in trades:
                        trades.pop(trade.order.id)
                    if not trades:
                        self.lmt_price_trades.pop(trade.contract.code)
                trade.status.status = sj.order.Status.Cancelled
        return {
            "operation": {"op_type": op_type, "op_code": op_code, "op_msg": ""},
            "order": {
                "id": trade.order.id,
                "seqno": trade.order.seqno,
                "ordno": trade.order.ordno,
                "action": trade.order.action,
                "price": trade.order.price,
                "quantity": trade.order.quantity,
                "order_cond": trade.order.order_cond,
                "order_lot": trade.order.order_lot,
                "custom_field": trade.order.custom_field,
                "order_type": trade.order.order_type,
                "price_type": trade.order.price_type,
            },
            "status": {
                "id": trade.order.id,
                "exchange_ts": datetime.datetime.now().timestamp(),
                "order_quantity": trade.order.quantity if op_type == "New" else 0,
                "modified_price": 0.0,
                "cancel_quantity": cancel_quantity if op_type == "Cancel" else 0,
                "web_id": "137",
            },
            "contract": {
                "security_type": trade.contract.security_type,
                "exchange": trade.contract.exchange,
                "code": trade.contract.code,
                "symbol": trade.contract.symbol,
                "name": trade.contract.name,
                "currency": trade.contract.currency,
            },
        }

    def gen_deal_msg(self, trade: sj.order.Trade, quantity: int, price: float):
        trade.status.deal_quantity += quantity
        trade.status.status = (
            sj.order.Status.Filled
            if quantity == trade.order.quantity
            else sj.order.Status.PartFilled
        )
        return {
            "trade_id": "12ab3456",
            "exchange_seq": "123456",
            "broker_id": "your_broker_id",
            "account_id": "your_account_id",
            "action": trade.order.action,
            "code": trade.contract.code,
            "order_cond": trade.order.order_cond,
            "order_lot": trade.order.order_lot,
            "price": price,
            "quantity": quantity,
            "web_id": "137",
            "custom_field": trade.order.custom_field,
            "ts": datetime.datetime.now().timestamp(),
        }

    def call_order_callback(self, trade: sj.order.Trade, op_type: str):
        time.sleep(0.5)
        order_msg = self.gen_order_msg(trade, op_type)
        self.order_callback(OrderState.TFTOrder, order_msg)
        time.sleep(0.1)
        if trade.order.price_type == TFTStockPriceType.MKT:
            if trade.status.status != sj.order.Status.Cancelled:
                s = self.snapshots.get(trade.contract.code)
                deal_msg = self.gen_deal_msg(
                    trade,
                    quantity=trade.order.quantity,
                    price=s.price if s else trade.order.price,
                )
                self.order_callback(OrderState.TFTDeal, deal_msg)
        else:
            if trade.status.status != sj.order.Status.Cancelled:
                with self.lock:
                    if trade.contract.code in self.lmt_price_trades:
                        self.lmt_price_trades[trade.contract.code][
                            trade.order.id
                        ] = trade
                    else:
                        self.lmt_price_trades[trade.contract.code] = {
                            trade.order.id: trade
                        }



import shioaji as sj
from typing import List
from threading import Lock
from dataclasses import dataclass, field
from shioaji.constant import (
    TFTStockPriceType,
)


@dataclass
class PriceSet:
    price: float
    quantity: int
    price_type: TFTStockPriceType
    in_transit_quantity: int = 0
    # time: datetime.time
    # time_cond: TimeCond


@dataclass
class PositionCond:
    quantity: int
    entry_price: List[PriceSet]
    stop_loss_price: List[PriceSet]
    stop_profit_price: List[PriceSet]
    cover_price: List[PriceSet] = field(default_factory=list)


@dataclass
class PositionStatus:
    cancel_preorder: bool = False
    cancel_quantity: int = 0
    entry_order_quantity: int = 0
    entry_quantity: int = 0
    open_quantity: int = 0
    cover_order_quantity: int = 0
    cover_quantity: int = 0


@dataclass
class Position:
    contract: sj.contracts.Contract
    cond: PositionCond
    status: PositionStatus = field(default_factory=PositionStatus)
    entry_trades: List[sj.order.Trade] = field(default_factory=list)
    cover_trades: List[sj.order.Trade] = field(default_factory=list)
    lock: Lock = field(default_factory=Lock)



""" sjtrade
trading with shioaji
"""

__version__ = "0.4.6"

def inject_env():
    import os

    if os.environ.get("LOGURU_FORMAT", None) is None:
        os.environ["LOGURU_FORMAT"] = (
            "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green>"
            " | <level>{level: <8}</level>"
            " | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>"
            " | <level>{message}</level>"
        )

inject_env()
from .trader import SJTrader
from .strategy import StrategyBase



import loguru
import pytest
import shioaji as sj
from pytest_mock import MockFixture


@pytest.fixture
def stock_contracts_raw():
    return [
        {
            "security_type": "STK",
            "exchange": "TSE",
            "code": "1605",
            "symbol": "TSE1605",
            "name": "華新",
            "currency": "TWD",
            "unit": 1000,
            "limit_up": 43.3,
            "limit_down": 35.5,
            "reference": 39.4,
            "update_date": "2022/05/19",
            "margin_trading_balance": 29297,
            "short_selling_balance": 1596,
            "day_trade": "Yes",
        },
        {
            "security_type": "STK",
            "exchange": "OTC",
            "code": "6290",
            "symbol": "OTC6290",
            "name": "良維",
            "currency": "TWD",
            "unit": 1000,
            "limit_up": 63.0,
            "limit_down": 51.6,
            "reference": 57.3,
            "update_date": "2022/05/19",
            "margin_trading_balance": 354,
            "short_selling_balance": 144,
            "day_trade": "Yes",
        },
    ]


@pytest.fixture
def api(mocker: MockFixture, stock_contracts_raw: list) -> sj.Shioaji:
    mock_api = mocker.MagicMock()
    mock_api.Contracts = sj.contracts.Contracts()
    stream_contracts = sj.contracts.StreamStockContracts(stock_contracts_raw)
    mock_api.Contracts.Stocks.append(stream_contracts)
    mock_api.Contracts.Indexs.set_status_fetched()
    mock_api.Contracts.Stocks.set_status_fetched()
    mock_api.Contracts.Futures.set_status_fetched()
    mock_api.Contracts.Options.set_status_fetched()
    mock_api.Contracts.status = sj.contracts.FetchStatus.Fetched
    return mock_api


@pytest.fixture
def api(mocker: MockFixture, stock_contracts_raw: list) -> sj.Shioaji:
    mock_api = mocker.MagicMock()
    mock_api.Contracts = sj.contracts.Contracts()
    stream_contracts = sj.contracts.StreamStockContracts(stock_contracts_raw)
    mock_api.Contracts.Stocks.append(stream_contracts)
    mock_api.Contracts.Indexs.set_status_fetched()
    mock_api.Contracts.Stocks.set_status_fetched()
    mock_api.Contracts.Futures.set_status_fetched()
    mock_api.Contracts.Options.set_status_fetched()
    mock_api.Contracts.status = sj.contracts.FetchStatus.Fetched
    return mock_api


@pytest.fixture
def logger(mocker: MockFixture) -> loguru._logger.Logger:
    return mocker.patch("sjtrade.trader.logger")


@pytest.fixture
def logger_stratagy(mocker: MockFixture) -> loguru._logger.Logger:
    return mocker.patch("sjtrade.strategy.logger")



import datetime
from typing import List
import pytest
from pytest_mock import MockerFixture
from sjtrade.utils import (
    price_between_tick,
    price_ceil,
    price_floor,
    price_move,
    price_round,
    quantity_num_split,
    quantity_split,
    sleep_until,
)


@pytest.mark.parametrize(
    ("input", "expected"),
    [
        (9.989, 9.99),
        (10.01, 10.05),
        (50.01, 50.1),
        (100.01, 100.5),
        (500.01, 501),
        (1000.01, 1005),
        (5000.01, 5010),
        (0.111, 0.12),
        (1.111, 1.12),
        (9.999, 10),
        (10.0377878, 10.05),
    ],
)
def test_price_ceil(input: float, expected: float):
    assert price_ceil(input) == expected


@pytest.mark.parametrize(
    ("input", "expected"),
    [
        (9.999, 9.99),
        (10.19, 10.15),
        (50.19, 50.1),
        (101.9, 101.5),
        (519.9, 519.0),
        (1019.9, 1015.0),
        (5199.9, 5190.0),
        (0.119, 0.11),
        (1.119, 1.11),
        (9.999, 9.99),
    ],
)
def test_price_floor(input: float, expected: float):
    assert price_floor(input) == expected


@pytest.mark.parametrize(
    ("price", "up", "expected"),
    [
        (9.999, False, 9.99),
        (10.19, False, 10.15),
        (50.19, False, 50.1),
        (101.9, False, 101.5),
        (519.9, False, 519.0),
        (9.989, True, 9.99),
        (10.01, True, 10.05),
        (50.01, True, 50.1),
        (100.01, True, 100.5),
        (500.01, True, 501),
        (34.05, False, 34.05),
        # TODO case 10% over limit up
    ],
)
def test_price_round(price: float, up: bool, expected: float):
    assert price_round(price, up) == expected

@pytest.mark.parametrize(
    ("price", "up", "expected"),
    [
        (9.9, 0, 9.9),
        (10.05, 1, 10.1),
        (10.05, 5, 10.3),
        (50, 1, 50.1),
        (49.9, 4, 50.2),
    ],
)
def test_price_move(price: float, up: bool, expected: float):
    assert price_move(price, up) == expected


@pytest.mark.parametrize(
    ("price0", "price1", "expected"),
    [
        (100, 102, 4),
        (102, 101, -2),
    ],
)
def test_price_between_tick(price0: float, price1: bool, expected: float):
    assert price_between_tick(price0, price1) == expected


@pytest.mark.parametrize(
    ("quantity", "num", "expected"),
    [
        (10 , 1, [10,]),
        (10 , 3, [4, 3, 3]),
        (8 , 3, [3, 3, 2]),
        (15 , 4, [4, 4, 4, 3]),
        (7 , 3, [3, 2, 2]),
        (-7 , 3, [-3, -2, -2]),
    ],
)
def test_quantity_num_split(quantity: int, num: int, expected: int):
    assert quantity_num_split(quantity, num) == expected


@pytest.mark.freeze_time("2022-06-06 00:30:00 UTC")
def test_sleep_until(mocker: MockerFixture):
    sleep_mock = mocker.patch("time.sleep")
    sleep_until(datetime.time(9, 0, 1))
    sleep_mock.assert_called_once_with(30 * 60 + 1)


@pytest.mark.freeze_time("2022-06-06 00:30:00 UTC")
def test_sleep_until(mocker: MockerFixture):
    sleep_mock = mocker.patch("time.sleep")
    sleep_until((9, 0, 1))
    sleep_mock.assert_called_once_with(30 * 60 + 1)


@pytest.mark.parametrize(
    ("quantity", "threshold", "expected"),
    [
        (
            497,
            499,
            [
                497,
            ],
        ),
        (
            503,
            499,
            [499, 4],
        ),
        (
            -1024,
            499,
            [-499, -499, -26],
        ),
    ],
)
def test_quantiy_split(quantity: int, threshold: int, expected: List[int]):
    res = quantity_split(quantity, threshold)
    assert res == expected



import pytest

from sjtrade.strategy import StrategyBase


def test_stratage_base():
    stratage = StrategyBase()
    with pytest.raises(NotImplementedError):
        stratage.entry_positions()

    with pytest.raises(NotImplementedError):
        stratage.cover_price_set(None)

    with pytest.raises(NotImplementedError):
        stratage.cover_positions(None)


    


import pytest
import time
import datetime
import loguru
from pytest_mock import MockFixture, MockerFixture
import shioaji as sj

from decimal import Decimal
from dataclasses import dataclass
from sjtrade.position import PriceSet
from sjtrade.trader import (
    Position,
    PositionCond,
    SJTrader,
    SimulationShioaji,
    StrategyBasic,
)
from shioaji.constant import (
    Action,
    TFTStockPriceType,
    TFTOrderType,
    QuoteVersion,
    Exchange,
    OrderState,
    TFTStockOrderLot,
    StockOrderCond,
    StockFirstSell,
)


@dataclass
class TickSTKv1:
    code: str
    datetime: datetime.datetime
    close: Decimal
    simtrade: bool


@pytest.fixture
def positions():
    return {"1605": -1, "6290": -3, "0000": 3}


@pytest.fixture
def sjtrader(api: sj.Shioaji, mocker: MockFixture, positions: dict) -> SJTrader:
    sjtrader = SJTrader(api)
    sjtrader.stratagy = StrategyBasic(entry_pct=0.05, contracts=api.Contracts)
    sjtrader.stratagy.read_position_func = mocker.MagicMock()
    sjtrader.stratagy.read_position_func.return_value = positions
    return sjtrader


@pytest.fixture
def sjtrader_sim(api: sj.Shioaji, mocker: MockFixture, positions: dict) -> SJTrader:
    sjtrader = SJTrader(api, simulation=True)
    sjtrader.stratagy = StrategyBasic(entry_pct=0.05, contracts=api.Contracts)
    sjtrader.stratagy.read_position_func = mocker.MagicMock()
    sjtrader.stratagy.read_position_func.return_value = positions
    return sjtrader


@pytest.fixture
def sjtrader_entryed(sjtrader: SJTrader):
    sjtrader.api.place_order.side_effect = lambda contract, order, timeout: (
        sj.order.Trade(
            contract,
            order,
            sj.order.OrderStatus(status=sj.order.Status.PreSubmitted),
        )
    )
    res = sjtrader.place_entry_positions()
    return sjtrader


@pytest.fixture
def sjtrader_entryed_sim(sjtrader_sim: SJTrader, positions: dict) -> SJTrader:
    res = sjtrader_sim.place_entry_positions()
    return sjtrader_sim


def test_sjtrader(api: sj.Shioaji):
    sjtrader = SJTrader(api)
    sjtrader.stratagy = StrategyBasic()
    assert hasattr(sjtrader, "api")
    sjtrader.stop_profit_pct = 0.1
    assert sjtrader.stop_profit_pct == 0.1
    assert sjtrader.stratagy.stop_profit_pct == 0.1
    sjtrader.stop_loss_pct = 0.1
    assert sjtrader.stop_loss_pct == 0.1
    assert sjtrader.stratagy.stop_loss_pct == 0.1
    sjtrader.position_filepath = "pos.txt"
    assert sjtrader.position_filepath == "pos.txt"
    assert sjtrader.stratagy.position_filepath == "pos.txt"
    sjtrader.entry_pct = 0.07
    assert sjtrader.entry_pct == 0.07
    assert sjtrader.stratagy.entry_pct == 0.07


def test_sjtrader_start(sjtrader: SJTrader, mocker: MockerFixture):
    sleep_until_mock = mocker.patch("sjtrade.trader.sleep_until")
    sjtrader.start()
    sjtrader.stratagy.read_position_func.assert_called_once()
    sjtrader.api.set_order_callback.assert_called_once_with(sjtrader.order_deal_handler)
    sjtrader.api.quote.set_on_tick_stk_v1_callback.assert_has_calls(
        [((sjtrader.cancel_preorder_handler,),), ((sjtrader.intraday_handler,),)]
    )
    sleep_until_mock.assert_has_calls(
        [
            mocker.call(datetime.time(8, 45)),
            mocker.call(datetime.time(8, 54, 59)),
            mocker.call(datetime.time(8, 59, 55)),
            mocker.call(datetime.time(13, 25, 59)),
        ]
    )


def test_sjtrader_sj_event_handler(sjtrader: SJTrader, logger: loguru._logger.Logger):
    sjtrader.sj_event_handel(0, 0, "info", "event")
    logger.info.assert_called_once()


def test_sjtrader_place_entry_order(
    sjtrader: SJTrader,
    logger: loguru._logger.Logger,
    logger_stratagy: loguru._logger.Logger,
):
    sjtrader.api.place_order.side_effect = (
        lambda contract, order, timeout: sj.order.Trade(
            contract, order, sj.order.OrderStatus(status=sj.order.Status.PreSubmitted)
        )
    )
    sjtrader.stop_loss_pct = 0.085
    sjtrader.stop_profit_pct = 0.09
    res = sjtrader.place_entry_positions()
    logger_stratagy.warning.assert_called_once()
    sjtrader.api.quote.subscribe.assert_has_calls(
        [
            ((sjtrader.api.Contracts.Stocks["1605"],), dict(version=QuoteVersion.v1)),
            ((sjtrader.api.Contracts.Stocks["6290"],), dict(version=QuoteVersion.v1)),
        ]
    )
    assert logger.info.call_count == 2
    sjtrader.api.update_status.assert_called_once()
    expected = [
        sj.order.Trade(
            sjtrader.api.Contracts.Stocks["1605"],
            sj.Order(
                price=41.35,
                quantity=1,
                action=Action.Sell,
                price_type=TFTStockPriceType.LMT,
                order_type=TFTOrderType.ROD,
                custom_field=sjtrader.name,
                first_sell=StockFirstSell.Yes,
            ),
            status=sj.order.OrderStatus(status=sj.order.Status.PreSubmitted),
        ),
        sj.order.Trade(
            sjtrader.api.Contracts.Stocks["6290"],
            sj.Order(
                price=60.1,
                quantity=3,
                action=Action.Sell,
                price_type=TFTStockPriceType.LMT,
                order_type=TFTOrderType.ROD,
                first_sell=StockFirstSell.Yes,
                custom_field=sjtrader.name,
            ),
            status=sj.order.OrderStatus(status=sj.order.Status.PreSubmitted),
        ),
    ]
    cond_1605 = PositionCond(
        quantity=-1,
        entry_price=[
            PriceSet(price=41.35, quantity=-1, price_type=TFTStockPriceType.LMT, in_transit_quantity=-1)
        ],
        stop_loss_price=[
            PriceSet(price=42.7, quantity=-1, price_type=TFTStockPriceType.MKT)
        ],
        stop_profit_price=[
            PriceSet(price=35.9, quantity=-1, price_type=TFTStockPriceType.MKT)
        ],
        cover_price=[],
    )
    cond_6290 = PositionCond(
        quantity=-3,
        entry_price=[
            PriceSet(price=60.1, quantity=-3, price_type=TFTStockPriceType.LMT, in_transit_quantity=-3)
        ],
        stop_loss_price=[
            PriceSet(price=62.1, quantity=-3, price_type=TFTStockPriceType.MKT)
        ],
        stop_profit_price=[
            PriceSet(price=52.2, quantity=-3, price_type=TFTStockPriceType.MKT)
        ],
        cover_price=[],
    )
    assert sjtrader.positions["1605"].cond == cond_1605
    assert sjtrader.positions["6290"].cond == cond_6290
    assert sjtrader.positions == {
        "1605": Position(
            contract=sjtrader.api.Contracts.Stocks["1605"],
            cond=cond_1605,
            entry_trades=[
                expected[0],
            ],
            lock=sjtrader.positions["1605"].lock
            # cover_trades=[],
        ),
        "6290": Position(
            contract=sjtrader.api.Contracts.Stocks["6290"],
            cond=cond_6290,
            entry_trades=[
                expected[1],
            ],
            lock=sjtrader.positions["6290"].lock
            # cover_trades=[],
        ),
    }
    assert len(res) == 2
    assert res == sjtrader.positions


def test_sjtrader_cancel_preorder_handler(
    sjtrader_entryed: SJTrader, mocker: MockerFixture, logger: loguru._logger.Logger
):
    tick = TickSTKv1("1605", "2022-05-25 08:45:01", Decimal("43.3"), True)

    def make_cancel_order_status(trade):
        trade.status.status = sj.order.Status.Cancelled
        trade.status.cancel_quantity = trade.order.quantity

    sjtrader_entryed.update_status = mocker.MagicMock(
        side_effect=make_cancel_order_status
    )

    sjtrader_entryed.cancel_preorder_handler(Exchange.TSE, tick)
    sjtrader_entryed.api.cancel_order.assert_called_once_with(
        sjtrader_entryed.positions["1605"].entry_trades[0], timeout=0
    )
    assert logger.info.called
    # TODO need single trade update status interface
    # sjtrader_entryed.update_status.assert_called_once_with(
    #     sjtrader_entryed.positions["1605"].entry_trades[0]
    # )
    # sjtrader.api._solace.update_status.assert_called()
    # assert sjtrader_entryed.positions["1605"].cancel_quantity == -1


def test_sjtrader_re_entry_order(
    sjtrader_entryed: SJTrader, mocker: MockerFixture, logger: loguru._logger.Logger
):
    tick = TickSTKv1("1605", "2022-05-25 08:45:01", Decimal("43.3"), True)

    def make_cancel_order_status(trade, timeout):
        trade.status.status = sj.order.Status.Cancelled
        trade.status.cancel_quantity = trade.order.quantity

    sjtrader_entryed.update_status = mocker.MagicMock(
        side_effect=make_cancel_order_status
    )
    position = sjtrader_entryed.positions["1605"]
    sjtrader_entryed.cancel_preorder_handler(position, tick)

    tick = TickSTKv1("1605", "2022-05-25 09:00:01", Decimal("35"), False)
    sjtrader_entryed.re_entry_order(position, tick)
    sjtrader_entryed.api.place_order.assert_called_with(
        contract=position.contract,
        order=sj.order.TFTStockOrder(
            price=0,
            quantity=1,
            action=Action.Sell,
            price_type=TFTStockPriceType.MKT,
            order_type=TFTOrderType.ROD,
            first_sell=StockFirstSell.Yes,
            custom_field="dt1",
        ),
        timeout=0,
    )
    assert logger.info.called


def test_sjtrader_stop_profit(
    sjtrader_entryed: SJTrader, mocker: MockerFixture, logger: loguru._logger.Logger
):
    tick = TickSTKv1("1605", "2022-05-25 08:45:01", Decimal("35.5"), False)
    position = sjtrader_entryed.positions["1605"]
    position.status.open_quantity = -1
    sjtrader_entryed.place_cover_order = mocker.MagicMock()
    sjtrader_entryed.stop_profit(position, tick)
    sjtrader_entryed.place_cover_order.assert_called_once()
    assert logger.info.called


def test_sjtrader_stop_loss(
    sjtrader_entryed: SJTrader, mocker: MockerFixture, logger: loguru._logger.Logger
):
    tick = TickSTKv1("1605", "2022-05-25 08:45:01", Decimal("43.3"), False)
    position = sjtrader_entryed.positions["1605"]
    position.status.open_quantity = -1
    sjtrader_entryed.place_cover_order = mocker.MagicMock()
    sjtrader_entryed.stop_loss(position, tick)
    sjtrader_entryed.place_cover_order.assert_called_once()
    assert logger.info.called


def test_sjtrader_intraday_handler(sjtrader_entryed: SJTrader, mocker: MockerFixture):
    tick = TickSTKv1("1605", "2022-05-25 08:45:01", Decimal("43.3"), True)
    sjtrader_entryed.re_entry_order = mocker.MagicMock()
    sjtrader_entryed.stop_profit = mocker.MagicMock()
    sjtrader_entryed.stop_loss = mocker.MagicMock()
    sjtrader_entryed.intraday_handler(Exchange.TSE, tick)
    sjtrader_entryed.re_entry_order.assert_called_once()
    sjtrader_entryed.stop_profit.assert_called_once()
    sjtrader_entryed.stop_loss.assert_called_once()


def test_sjtrader_place_cover_order(
    sjtrader_entryed: SJTrader, logger: loguru._logger.Logger
):
    position = sjtrader_entryed.positions["1605"]
    position.status.open_quantity = -1
    sjtrader_entryed.place_cover_order(position)
    order_msg = gen_sample_order_msg("1605", Action.Buy, 1, op_type="New", op_code="00")
    sjtrader_entryed.order_handler(order_msg, position)
    assert position.status.cover_order_quantity == 1
    position.cover_trades[0].order.price == 43.3
    # TODO assert called once with
    # sjtrader_entryed.api.place_order.assert_has_calls([
    #     ((), dict(contract=position.contract, order=sj.order.TFTStockOrder(
    #         price=0,
    #         quantity=
    #     )))
    # ])
    assert len(position.cover_trades) == 1
    assert logger.info.called


def test_sjtrader_open_position_cover(
    sjtrader_entryed: SJTrader, logger: loguru._logger.Logger
):
    position = sjtrader_entryed.positions["1605"]
    position.open_quantity = -1
    sjtrader_entryed.place_cover_order(position)
    order_msg = gen_sample_order_msg("1605", Action.Buy, 1, op_type="New", op_code="00")
    sjtrader_entryed.order_handler(order_msg, position)
    sjtrader_entryed.open_position_cover()
    assert logger.info.called


@pytest.fixture
def order_msg():
    return {
        "operation": {"op_type": "New", "op_code": "00", "op_msg": ""},
        "order": {
            "id": "c21b876d",
            "seqno": "429832",
            "ordno": "W2892",
            "action": "Sell",
            "price": 44.3,
            "quantity": 1,
            "order_cond": "Cash",
            "order_lot": "Common",
            "custom_field": "dt1",
            "order_type": "ROD",
            "price_type": "LMT",
        },
        "status": {
            "id": "c21b876d",
            "exchange_ts": 1583828972,
            "modified_price": 0,
            "cancel_quantity": 0,
            "web_id": "137",
        },
        "contract": {
            "security_type": "STK",
            "exchange": "TSE",
            "code": "1605",
            "symbol": "",
            "name": "",
            "currency": "TWD",
        },
    }


@pytest.fixture
def deal_msg():
    return {
        "trade_id": "12ab3456",
        "exchange_seq": "123456",
        "broker_id": "your_broker_id",
        "account_id": "your_account_id",
        "action": Action.Buy,
        "code": "1605",
        "order_cond": StockOrderCond.Cash,
        "order_lot": TFTStockOrderLot.Common,
        "price": 12,
        "quantity": 10,
        "web_id": "137",
        "custom_field": "dt1",
        "ts": 1583828972,
    }


def gen_sample_order_msg(
    code: str, action: Action, quantity: int, op_type: str, op_code: str
):
    return {
        "operation": {"op_type": op_type, "op_code": op_code, "op_msg": ""},
        "order": {
            "id": "c21b876d",
            "seqno": "429832",
            "ordno": "W2892",
            "action": action,
            "price": 44.3,
            "quantity": quantity,
            "order_cond": "Cash",
            "order_lot": "Common",
            "custom_field": "dt1",
            "order_type": "ROD",
            "price_type": "LMT",
        },
        "status": {
            "id": "c21b876d",
            "exchange_ts": 1583828972,
            "order_quantity": quantity if op_type == "New" else 0,
            "modified_price": 0.0,
            "cancel_quantity": quantity if op_type == "Cancel" else 0,
            "web_id": "137",
        },
        "contract": {
            "security_type": "STK",
            "exchange": "TSE",
            "code": code,
            "symbol": "",
            "name": "",
            "currency": "TWD",
        },
    }


def gen_sample_deal_msg(code: str, action: Action, quantity: int):
    return {
        "trade_id": "12ab3456",
        "exchange_seq": "123456",
        "broker_id": "your_broker_id",
        "account_id": "your_account_id",
        "action": action,
        "code": code,
        "order_cond": StockOrderCond.Cash,
        "order_lot": TFTStockOrderLot.Common,
        "price": 12,
        "quantity": quantity,
        "web_id": "137",
        "custom_field": "dt1",
        "ts": 1583828972,
    }


def test_sjtrader_order_deal_handler_receive_order_msg(
    sjtrader_entryed: SJTrader, order_msg: dict, mocker: MockerFixture
):
    sjtrader_entryed.order_handler = mocker.MagicMock()
    sjtrader_entryed.order_deal_handler(OrderState.TFTOrder, order_msg)
    sjtrader_entryed.order_handler.assert_called_once_with(
        order_msg,
        sjtrader_entryed.positions["1605"],
    )


def test_sjtrader_order_deal_handler_receive_deal_msg(
    sjtrader_entryed: SJTrader, deal_msg: dict, mocker: MockerFixture
):
    sjtrader_entryed.deal_handler = mocker.MagicMock()
    sjtrader_entryed.order_deal_handler(OrderState.TFTDeal, deal_msg)
    sjtrader_entryed.deal_handler.assert_called_once_with(
        deal_msg,
        sjtrader_entryed.positions["1605"],
    )


def test_sjtrader_order_handler(
    sjtrader_entryed: SJTrader, logger: loguru._logger.Logger
):
    position = sjtrader_entryed.positions["1605"]
    order_msg = gen_sample_order_msg(
        "1605", Action.Sell, 1, op_type="New", op_code="00"
    )
    sjtrader_entryed.order_handler(order_msg, position)
    assert position.status.entry_order_quantity == -1

    order_msg = gen_sample_order_msg(
        "1605", Action.Sell, 1, op_type="Cancel", op_code="00"
    )
    sjtrader_entryed.order_handler(order_msg, position)
    assert position.status.entry_order_quantity == 0

    order_msg = gen_sample_order_msg("1605", Action.Buy, 1, op_type="New", op_code="00")
    sjtrader_entryed.order_handler(order_msg, position)
    assert position.status.cover_order_quantity == 1

    order_msg = gen_sample_order_msg(
        "1605", Action.Buy, 1, op_type="Cancel", op_code="00"
    )
    sjtrader_entryed.order_handler(order_msg, position)
    assert position.status.cover_order_quantity == 0
    assert logger.info.called


def test_sjtrader_deal_handler(
    sjtrader_entryed: SJTrader, logger: loguru._logger.Logger
):
    position = sjtrader_entryed.positions["1605"]
    deal_msg = gen_sample_deal_msg("1605", Action.Sell, 1)
    sjtrader_entryed.deal_handler(deal_msg, position)
    assert position.status.open_quantity == -1
    assert position.status.entry_quantity == -1
    deal_msg = gen_sample_deal_msg("1605", Action.Buy, 1)
    sjtrader_entryed.deal_handler(deal_msg, position)
    assert position.status.open_quantity == 0
    assert position.status.cover_quantity == 1
    assert logger.info.called


def test_sim_sj_quote_callback(api: sj.Shioaji):
    sim_api = SimulationShioaji(print)
    contract = api.Contracts.Stocks["1605"]
    order = sj.Order(
        price=35,
        quantity=1,
        action=Action.Sell,
        price_type=TFTStockPriceType.LMT,
        order_type=TFTOrderType.ROD,
        custom_field="dt1",
    )
    sim_api.place_order(
        contract,
        order,
    )
    order = sj.Order(
        price=34,
        quantity=1,
        action=Action.Sell,
        price_type=TFTStockPriceType.LMT,
        order_type=TFTOrderType.ROD,
        custom_field="dt1",
    )
    sim_api.place_order(
        contract,
        order,
    )
    time.sleep(1.1)
    tick = TickSTKv1("1605", "2022-05-25 09:00:01", 34.5, False)
    assert len(sim_api.lmt_price_trades["1605"]) == 2
    sim_api.quote_callback(Exchange.TSE, tick)
    assert sim_api.snapshots["1605"].price == 34.5
    assert len(sim_api.lmt_price_trades["1605"]) == 1
    tick = TickSTKv1("1605", "2022-05-25 09:00:02", 36, False)
    sim_api.quote_callback(Exchange.TSE, tick)
    assert sim_api.snapshots["1605"].price == 36
    assert "1605" not in sim_api.lmt_price_trades


def test_sim_sj_place_order(api: sj.Shioaji):
    sim_api = SimulationShioaji(print)
    contract = api.Contracts.Stocks["1605"]
    order = sj.Order(
        price=41.35,
        quantity=1,
        action=Action.Sell,
        price_type=TFTStockPriceType.MKT,
        order_type=TFTOrderType.ROD,
        custom_field="dt1",
    )
    trade = sim_api.place_order(
        contract,
        order,
    )
    time.sleep(0.55)
    assert trade.contract == contract
    # assert trade.order == order
    assert trade.order.seqno == "000001"
    assert trade.order.id == "c2c49415"
    assert trade.order.ordno != ""
    assert trade.status.status == "Submitted"
    time.sleep(0.1)
    assert trade.status.deal_quantity == 1


def test_sim_sj_cancel_order(api: sj.Shioaji):
    sim_api = SimulationShioaji(print)
    contract = api.Contracts.Stocks["1605"]
    order = sj.Order(
        price=41.35,
        quantity=1,
        action=Action.Sell,
        price_type=TFTStockPriceType.LMT,
        order_type=TFTOrderType.ROD,
        custom_field="dt1",
    )
    trade = sim_api.place_order(
        contract,
        order,
    )
    time.sleep(0.55)
    trade = sim_api.cancel_order(trade)
    time.sleep(0.55)
    assert trade.status.status == sj.order.Status.Cancelled

    order = sj.Order(
        price=41.35,
        quantity=1,
        action=Action.Sell,
        price_type=TFTStockPriceType.MKT,
        order_type=TFTOrderType.ROD,
        custom_field="dt1",
    )
    trade = sim_api.place_order(
        contract,
        order,
    )
    time.sleep(0.55)
    trade = sim_api.cancel_order(trade)
    time.sleep(0.55)
    assert trade.status.status == sj.order.Status.Filled


def test_sim_sj_update_status(api: sj.Shioaji):
    sim_api = SimulationShioaji(print)
    sim_api.update_status()


def test_sjtrader_sim(api: sj.Shioaji):
    sjtrader = SJTrader(api, simulation=True)
    assert hasattr(sjtrader, "simulation_api")


def test_sjtrader_sim_place_entry_order(
    sjtrader_sim: SJTrader,
    logger: loguru._logger.Logger,
    logger_stratagy: loguru._logger.Logger,
):
    res = sjtrader_sim.place_entry_positions()
    logger_stratagy.warning.assert_called_once()
    sjtrader_sim.api.quote.subscribe.assert_has_calls(
        [
            (
                (sjtrader_sim.api.Contracts.Stocks["1605"],),
                dict(version=QuoteVersion.v1),
            ),
            (
                (sjtrader_sim.api.Contracts.Stocks["6290"],),
                dict(version=QuoteVersion.v1),
            ),
        ]
    )
    assert logger.info.call_count == 2
    assert len(res) == 2


def test_sjtrader_sim_cancel_preorder_handler(
    sjtrader_entryed_sim: SJTrader, logger: loguru._logger.Logger
):
    tick = TickSTKv1("1605", "2022-05-25 08:45:01", 43.3, True)
    sjtrader_entryed_sim.cancel_preorder_handler(Exchange.TSE, tick)

    assert logger.info.called
    time.sleep(0.55)
    assert sjtrader_entryed_sim.positions["1605"].status.cancel_quantity == 1


def test_sjtrader_sim_re_entry_order(
    sjtrader_entryed_sim: SJTrader, logger: loguru._logger.Logger
):
    tick = TickSTKv1("1605", "2022-05-25 08:45:01", 43.3, True)
    position = sjtrader_entryed_sim.positions["1605"]
    sjtrader_entryed_sim.cancel_preorder_handler(position, tick)
    time.sleep(0.55)
    assert sjtrader_entryed_sim.positions["1605"].status.cancel_preorder == True
    tick = TickSTKv1("1605", "2022-05-25 09:00:01", 35, False)
    sjtrader_entryed_sim.re_entry_order(position, tick)
    assert logger.info.called
    time.sleep(0.65)
    assert sjtrader_entryed_sim.positions["1605"].status.entry_order_quantity == -1



import time
import datetime
import operator
from typing import Callable, Dict, List, Optional, Union
from concurrent.futures import Future, ThreadPoolExecutor
import shioaji as sj

from .utils import quantity_split, sleep_until
from .data import Snapshot
from .simulation_shioaji import SimulationShioaji
from .strategy import StrategyBasic
from .position import Position, PositionCond, PriceSet
from loguru import logger
from shioaji.constant import (
    Action,
    TFTStockPriceType,
    TFTOrderType,
    QuoteVersion,
    Exchange,
    OrderState,
    StockFirstSell,
)


logger.add("sjtrader.log", rotation="1 days")


class SJTrader:
    def __init__(self, api: sj.Shioaji, simulation: bool = False):
        self.api = api
        self.positions: Dict[str, Position] = {}
        self.snapshots: Dict[str, Snapshot] = {}
        self._stop_loss_pct = 0.09
        self._stop_profit_pct = 0.09
        self._entry_pct = 0.05
        self.open_price = {}
        self._position_filepath = "position.txt"
        self.name = "dt1"
        self.executor = ThreadPoolExecutor()
        self.simulation = simulation
        if simulation:
            self.simulation_api = SimulationShioaji(self.order_deal_handler)
        self.api.set_order_callback(self.order_deal_handler)
        self.api.quote.set_event_callback(self.sj_event_handel)
        self.stratagy = StrategyBasic(contracts=self.api.Contracts)
        # self.account = api.stock_account
        # self.entry_trades: Dict[str, sj.order.Trade] = {}

    def start(
        self,
        entry_time: datetime.time = datetime.time(8, 45),
        cancel_preorder_time: datetime.time = datetime.time(8, 54, 59),
        intraday_handler_time: datetime.time = datetime.time(8, 59, 55),
        cover_time: datetime.time = datetime.time(13, 25, 59),
    ):
        self.set_on_tick_handler(self.update_snapshot)
        entry_future = self.executor_on_time(entry_time, self.place_entry_positions)
        self.executor_on_time(
            cancel_preorder_time,
            self.set_on_tick_handler,
            self.cancel_preorder_handler,
        )
        self.executor_on_time(
            intraday_handler_time,
            self.set_on_tick_handler,
            self.intraday_handler,
        )
        self.executor_on_time(cover_time, self.open_position_cover)
        return entry_future

    def run_at(self, t: Union[datetime.time, tuple], func: Callable, *args, **kwargs):
        sleep_until(t)
        return func(*args, **kwargs)

    def executor_on_time(
        self, t: Union[datetime.time, tuple], func: Callable, *args, **kwargs
    ) -> Future:
        return self.executor.submit(self.run_at, t, func, *args, **kwargs)

    def set_on_tick_handler(self, func: Callable[[Exchange, sj.TickSTKv1], None]):
        self.api.quote.set_on_tick_stk_v1_callback(func)

    @property
    def stop_loss_pct(self) -> float:
        return self._stop_loss_pct

    @stop_loss_pct.setter
    def stop_loss_pct(self, v: float) -> float:
        self._stop_loss_pct = v
        self.stratagy.stop_loss_pct = self._stop_loss_pct

    @property
    def stop_profit_pct(self) -> float:
        return self._stop_profit_pct

    @stop_profit_pct.setter
    def stop_profit_pct(self, v: float) -> float:
        self._stop_profit_pct = v
        self.stratagy.stop_profit_pct = self._stop_profit_pct

    @property
    def entry_pct(self) -> float:
        return self._entry_pct

    @entry_pct.setter
    def entry_pct(self, v: float) -> float:
        self._entry_pct = v
        self.stratagy.entry_pct = self._entry_pct

    @property
    def position_filepath(self) -> float:
        return self._position_filepath

    @position_filepath.setter
    def position_filepath(self, v: str) -> float:
        self._position_filepath = v
        self.stratagy.position_filepath = self._position_filepath

    def sj_event_handel(self, resp_code: int, event_code: int, info: str, event: str):
        logger.info(
            f"Response Code: {resp_code} | Event Code: {event_code} | Info: {info} | Event: {event}"
        )

    def place_entry_order(
        self,
        code: str,
        pos: int,
        entry_price: List[PriceSet],
        stop_profit_price: List[PriceSet],
        stop_loss_price: List[PriceSet],
    ):
        api = self.simulation_api if self.simulation else self.api
        contract = self.api.Contracts.Stocks[code]
        if not contract:
            logger.warning(f"Code: {code} not exist in TW Stock.")
        else:
            position = self.positions[code] = Position(
                contract=contract,
                cond=PositionCond(
                    quantity=pos,
                    entry_price=entry_price,
                    stop_loss_price=stop_loss_price,
                    stop_profit_price=stop_profit_price,
                    cover_price=[],
                ),
            )
            self.snapshots[code] = Snapshot(price=0.0)
            self.api.quote.subscribe(contract, version=QuoteVersion.v1)
            for price_set in position.cond.entry_price:
                if abs(price_set.quantity) == abs(price_set.in_transit_quantity):
                    continue
                price, price_quantity = price_set.price, price_set.quantity
                quantity_s = quantity_split(price_quantity, threshold=499)
                with position.lock:
                    for q in quantity_s:
                        trade = api.place_order(
                            contract=contract,
                            order=sj.Order(
                                price=price,
                                quantity=abs(q),
                                action=Action.Buy if pos > 0 else Action.Sell,
                                price_type=price_set.price_type,
                                order_type=TFTOrderType.ROD,
                                first_sell=StockFirstSell.No
                                if pos > 0
                                else StockFirstSell.Yes,
                                custom_field=self.name,
                            ),
                            timeout=0,
                        )
                        price_set.in_transit_quantity += q
                        # position.status.entry_order_in_transit += q
                        position.entry_trades.append(trade)
                        logger.info(f"{code} | {trade.order}")

    def place_entry_positions(self) -> Dict[str, Position]:
        api = self.simulation_api if self.simulation else self.api
        for entry_kwarg in self.stratagy.entry_positions():
            self.place_entry_order(**entry_kwarg)
        api.update_status()
        return self.positions

    def update_snapshot(self, exchange: Exchange, tick: sj.TickSTKv1):
        self.snapshots[tick.code].price = tick.close

    def cancel_preorder_handler(self, exchange: Exchange, tick: sj.TickSTKv1):
        position = self.positions[tick.code]
        if self.simulation:
            api = self.simulation_api
        else:
            api = self.api
        # 8:55 - 8:59:55
        if tick.simtrade:
            if position.cond.quantity < 0 and float(tick.close) == position.contract.limit_up:
                with position.lock:
                    position.status.cancel_preorder = True
                for trade in self.positions[tick.code].entry_trades:
                    if trade.status.status != sj.order.Status.Cancelled:
                        api.cancel_order(trade, timeout=0)
                        logger.info(f"{trade.contract.code} | {trade.order}")
                        # check handle
                        # self.update_status(trade)
                        # if trade.status.status == sj.order.Status.Cancelled:
                        #     position.cancel_preorder = True
                        # else:
                        #     logger.warning("position {} not cancel....")
                        # TODO handel it

    def re_entry_order(self, position: Position, tick: sj.TickSTKv1):
        # 9:00 -> first
        if self.simulation:
            api = self.simulation_api
        else:
            api = self.api
        if not tick.simtrade:
            if tick.code not in self.open_price:
                self.open_price[tick.code] = tick.close
                if (
                    position.status.cancel_preorder
                    and float(tick.close) < position.cond.stop_loss_price[0].price
                ):  # TODO check min or max
                    trade = api.place_order(
                        contract=position.contract,
                        order=sj.order.TFTStockOrder(
                            price=0,
                            quantity=abs(position.cond.quantity),
                            action=Action.Buy
                            if position.cond.quantity > 0
                            else Action.Sell,
                            price_type=TFTStockPriceType.MKT,
                            order_type=TFTOrderType.ROD,
                            first_sell=StockFirstSell.No
                            if position.cond.quantity > 0
                            else StockFirstSell.Yes,
                            custom_field=self.name,
                        ),
                        timeout=0,
                    )
                    logger.info(f"{trade.contract.code} | {trade.order}")
                    api.update_status(trade=trade)

    def intraday_handler(self, exchange: Exchange, tick: sj.TickSTKv1):
        if self.simulation:
            self.simulation_api.quote_callback(exchange, tick)
        position = self.positions[tick.code]
        self.re_entry_order(position, tick)
        self.update_snapshot(exchange, tick)
        # 9:00 -> 13:24:49 stop loss stop profit
        self.stop_loss(position, tick)
        self.stop_profit(position, tick)

    def stop_profit(self, position: Position, tick: sj.TickSTKv1):
        if not tick.simtrade:
            cover_quantity = (
                position.status.open_quantity + position.status.cover_order_quantity
            )
            if cover_quantity == 0:
                return
            if position.status.open_quantity > 0:
                op = operator.ge
                cross = "over"
            else:
                op = operator.le
                cross = "under"
            for price_set in position.cond.stop_profit_price:
                if op(float(tick.close), price_set.price):
                    if abs(price_set.quantity) == abs(price_set.in_transit_quantity):
                        continue
                    self.place_cover_order(position, [price_set])
                    logger.info(
                        f"{position.contract.code} | price: {tick.close} cross {cross} {price_set.price} "
                        f"cover quantity: {price_set.quantity}"
                    )

    def stop_loss(self, position: Position, tick: sj.TickSTKv1):
        if not tick.simtrade:
            cover_quantity = (
                position.status.open_quantity + position.status.cover_order_quantity
            )
            if cover_quantity == 0:
                return
            if position.status.open_quantity > 0:
                op = operator.le
                cross = "under"
            else:
                op = operator.ge
                cross = "over"
            for price_set in position.cond.stop_loss_price:
                if op(float(tick.close), price_set.price):
                    if abs(price_set.quantity) == abs(price_set.in_transit_quantity):
                        continue
                    self.place_cover_order(position, [price_set])
                    logger.info(
                        f"{position.contract.code} | price: {tick.close} cross {cross} {price_set.price} "
                        f"cover quantity: {price_set.quantity}"
                    )

    def place_cover_order(
        self, position: Position, price_sets: List[PriceSet] = []
    ):  # TODO with price quantity
        if self.simulation:
            api = self.simulation_api
        else:
            api = self.api
        cover_quantity = (
            position.status.open_quantity + position.status.cover_order_quantity
        )
        if not price_sets:
            price_sets = self.stratagy.cover_price_set(
                position, self.snapshots[position.contract.code]
            )
            position.cond.cover_price += price_sets
        if cover_quantity == 0:
            return
        for price_set in price_sets:
            if abs(price_set.quantity) == abs(price_set.in_transit_quantity):
                continue
            if price_set.quantity:
                quantity_s = quantity_split(price_set.quantity, threshold=499)
                for q in quantity_s:
                    trade = api.place_order(
                        contract=position.contract,
                        order=sj.order.TFTStockOrder(
                            price=price_set.price,
                            quantity=abs(q),
                            action=Action.Buy
                            if position.cond.quantity < 0
                            else Action.Sell,
                            price_type=price_set.price_type,
                            order_type=TFTOrderType.ROD,
                            custom_field=self.name,
                        ),
                        timeout=0,
                    )
                    logger.info(f"{trade.contract.code} | {trade.order}")
                    price_set.in_transit_quantity += q
                    position.cover_trades.append(trade)
                    # api.update_status(trade=trade)

    def open_position_cover(self, onclose: bool = True):
        if self.simulation:
            api = self.simulation_api
        else:
            api = self.api
        api.update_status()
        logger.info(f"start place cover order. onclose: {onclose}")
        for code, position in self.positions.items():
            if position.status.cover_order_quantity and (
                position.status.cover_order_quantity != position.status.cover_quantity
            ):
                for trade in position.cover_trades:
                    if trade.status.status in [
                        sj.order.Status.Submitted,
                        sj.order.Status.PreSubmitted,
                        sj.order.Status.PartFilled,
                    ]:
                        api.cancel_order(trade, timeout=0)
            if position.status.entry_order_quantity and (
                position.status.entry_order_quantity != position.status.entry_quantity
            ):
                for trade in position.entry_trades:
                    if trade.status.status in [
                        sj.order.Status.Submitted,
                        sj.order.Status.PreSubmitted,
                        sj.order.Status.PartFilled,
                    ]:
                        api.cancel_order(trade, timeout=0)
            # event wait cancel
        for code, position in self.positions.items():
            for _ in range(10):
                if (
                    position.status.cover_quantity
                    != position.status.cover_order_quantity
                ):
                    time.sleep(1)
            if position.status.cover_quantity != position.status.cover_order_quantity:
                logger.error(
                    f"{code} | cancel not work, position cover order "
                    f"{position.status.cover_order_quantity}, position cover {position.status.cover_quantity}"
                )
        if onclose:
            self.positions = self.stratagy.cover_positions_onclose(self.positions)
        else:
            self.positions = self.stratagy.cover_positions(
                self.positions, self.snapshots
            )
        for code, position in self.positions.items():
            if position.status.open_quantity:
                self.place_cover_order(position, position.cond.cover_price)

    def order_deal_handler(self, order_stats: OrderState, msg: Dict):
        if (
            order_stats == OrderState.TFTOrder
            and msg["order"]["custom_field"] == self.name
        ):
            self.order_handler(msg, self.positions[msg["contract"]["code"]])
        elif order_stats == OrderState.TFTDeal and msg["custom_field"] == self.name:
            self.deal_handler(msg, self.positions[msg["code"]])

    def order_handler(self, msg: Dict, position: Position):
        if msg["operation"]["op_code"] == "00":
            with position.lock:
                if msg["operation"]["op_type"] == "New":
                    order_quantity = msg["status"].get("order_quantity", 0)
                    order_pirce = msg["order"].get("price", 0)
                    if msg["order"]["action"] == Action.Sell:
                        if position.cond.quantity < 0:
                            position.status.entry_order_quantity -= order_quantity
                            logger.info(
                                f"{position.contract.code} | place short entry order with quantity {order_quantity}, price: {order_pirce}"
                            )
                            logger.debug(
                                f"{position.contract.code} | {position.status}"
                            )
                        else:
                            position.status.cover_order_quantity -= order_quantity
                            logger.info(
                                f"{position.contract.code} | place long cover order with quantity {order_quantity}, price: {order_pirce}"
                            )
                            logger.debug(
                                f"{position.contract.code} | {position.status}"
                            )
                    else:
                        if position.cond.quantity < 0:
                            position.status.cover_order_quantity += order_quantity
                            logger.info(
                                f"{position.contract.code} | place short cover order with quantity {order_quantity}, price: {order_pirce}"
                            )
                            logger.debug(
                                f"{position.contract.code} | {position.status}"
                            )
                        else:
                            position.status.entry_order_quantity += order_quantity
                            logger.info(
                                f"{position.contract.code} | place long entry order with quantity {order_quantity}, price: {order_pirce}"
                            )
                            logger.debug(
                                f"{position.contract.code} | {position.status}"
                            )
                else:
                    cancel_quantity = msg["status"].get("cancel_quantity", 0)
                    if msg["order"]["action"] == Action.Sell:
                        if position.cond.quantity < 0:
                            position.status.entry_order_quantity += cancel_quantity
                            logger.info(
                                f"{position.contract.code} | canceled short entry order with {cancel_quantity}"
                            )
                            logger.debug(
                                f"{position.contract.code} | {position.status}"
                            )
                        else:
                            position.status.cover_order_quantity += cancel_quantity
                            logger.info(
                                f"{position.contract.code} | canceled long cover order with {cancel_quantity}"
                            )
                            logger.debug(
                                f"{position.contract.code} | {position.status}"
                            )
                    else:
                        if position.cond.quantity < 0:
                            position.status.cover_order_quantity -= cancel_quantity
                            logger.info(
                                f"{position.contract.code} | canel short cover order with {cancel_quantity}"
                            )
                            logger.debug(
                                f"{position.contract.code} | {position.status}"
                            )
                        else:
                            position.status.entry_order_quantity -= cancel_quantity
                            logger.info(
                                f"{position.contract.code} | canel long entry order with {cancel_quantity}"
                            )
                            logger.debug(
                                f"{position.contract.code} | {position.status}"
                            )
                    position.status.cancel_quantity += cancel_quantity
        else:
            logger.error(f"Please Check: {msg}")

    def deal_handler(self, msg: Dict, position: Position):
        with position.lock:
            deal_quantity = msg["quantity"]
            deal_price = msg["price"]
            if msg["action"] == Action.Sell:
                position.status.open_quantity -= deal_quantity
                if position.cond.quantity < 0:
                    position.status.entry_quantity -= deal_quantity
                    logger.info(
                        f"{position.contract.code} | short entry order deal with {deal_quantity}, price: {deal_price}"
                    )
                    logger.debug(f"{position.contract.code} | {position.status}")
                else:
                    position.status.cover_quantity -= deal_quantity
                    logger.info(
                        f"{position.contract.code} | long cover order deal with {deal_quantity}, price: {deal_price}"
                    )
                    logger.debug(f"{position.contract.code} | {position.status}")
            else:
                position.status.open_quantity += deal_quantity
                if position.cond.quantity < 0:
                    position.status.cover_quantity += deal_quantity
                    logger.info(
                        f"{position.contract.code} | short cover order deal with {deal_quantity}, price: {deal_price}"
                    )
                    logger.debug(f"{position.contract.code} | {position.status}")
                else:
                    position.status.entry_quantity += deal_quantity
                    logger.info(
                        f"{position.contract.code} | long entry order deal with {deal_quantity}, price: {deal_price}"
                    )
                    logger.debug(f"{position.contract.code} | {position.status}")



import pytest
from pytest_mock import MockFixture
from sjtrade.io.file import read_csv_position, read_position


def test_read_position(mocker: MockFixture):
    mocker.patch("pathlib.Path.is_file").return_value = True
    mocker.patch("pathlib.Path.exists").return_value = True
    mocker.patch(
        "pathlib.Path.read_text"
    ).return_value = "1524\t18.0\n2359\t10.0\n3141\t2.0\n3265\t6.0\n4133\t6.0\n5608\t6.0\n6104\t1.0\n6470\t4.0\n"
    res = read_position("position.txt")
    assert res == {
        "1524": 18,
        "2359": 10,
        "3141": 2,
        "3265": 6,
        "4133": 6,
        "5608": 6,
        "6104": 1,
        "6470": 4,
    }


def test_read_position_notfile():
    with pytest.raises(FileNotFoundError):
        read_position("")


def test_read_csv_position(mocker: MockFixture):
    mocker.patch("pathlib.Path.is_file").return_value = True
    mocker.patch("pathlib.Path.exists").return_value = True
    mocker.patch(
        "pathlib.Path.read_text"
    ).return_value = (
        "標的,張數,停損檔數,尾盤鋪單%數\n1319,-8,3,1\n1539,-9,3,1\n1760,-9,4,2\n1795,-9,2,1\n"
    )
    res = read_csv_position("position.csv")
    assert res == {
        "1319": {"pos": -8, "stop_loss_tick": 3, "cover_pct": 1},
        "1539": {"pos": -9, "stop_loss_tick": 3, "cover_pct": 1},
        "1760": {"pos": -9, "stop_loss_tick": 4, "cover_pct": 2},
        "1795": {"pos": -9, "stop_loss_tick": 2, "cover_pct": 1},
    }


def test_read_position_notfile():
    with pytest.raises(FileNotFoundError):
        read_csv_position("")
